#!/usr/bin/env python3

from sysrepo_api import sysrepo_api3 as api
import ucpeif.wireguardvpnif as wireguardif
import ucpeif.wg_concentratorif as wgc
import vppapi_if.vppapi_if as vppif
import edgility_db.edgility_db as db
from pyroute2 import IPDB
from tveapi.common import utils
import ipaddress

STATIC_POOL_NAME = 'static_clients_'
DYNAMIC_POOL_NAME = 'dynamic_clients_'

class SysrepoWireGuardConcentrator(api.SysrepoAPI):
    def __init__(self,
                 app_name="wg-concentrator",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 journal_log=True,
                 namespace="/ucpe:config/wg-concentrators:wg-concentrators",
                 indexes={"wg-concentrator": "name", "permit": "id", "deny": "id"},
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            journal_log=journal_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority
        )

        self._wgc_interfaces = []
    
    def fix_string(s):
        if isinstance(s, str):
            return(s)
        return s.decode('ascii').rstrip('\0')

    def get_startup_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("startup", full_config=full_config, full_path=full_path)
    
    def get_running_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("running", full_config=full_config, full_path=full_path)
    
    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")
    
    def is_vpp_running(self):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return
    
    def get_wireguard_concentrators(self, startup=False):
        config_indexed = {}
        sys_conf = self.get_startup_config(full_path=False) if startup else self.get_running_config(full_path=False)
        wg_concentrators = sys_conf.get('wg-concentrator', [])
        for concentrator_info in wg_concentrators:
            config_indexed[concentrator_info['name']] = concentrator_info
        return config_indexed
    
    def apply_startup_config(self):
        self.is_vpp_running()
        
        self.logger.debug('Apply STARTUP config for Wireguard Concentrator')
        config_indexed = self.get_wireguard_concentrators(startup=True)
        for conc_int_name, config in sorted(config_indexed.items()):
            self.logger.debug(f"startup conc_int_name {conc_int_name}")
            concentrator_endpoint = config.get('concentrator-endpoint', '')
            concentrator_port = config.get('concentrator-port', None)
            concentrator_password = config.get('concentrator-password', '')
            concentrator_tunnel_ip = config.get('ipv4-prefix', '')
            is_enabled = config.get('enable', False)
            domain_name = config.get('domain-name', 'edgilityconcentrator.com')
            access_list = config.get('access-list', None)

            wgc_if = wireguardif.WireGuardVPNInterface(conc_int_name, logger=self.logger, l_port=concentrator_port,
                                                l_endpoint=concentrator_endpoint, enable=is_enabled, is_concentrator=True, d_name=domain_name, a_list=access_list)
            self._wgc_interfaces.append(wgc_if)
            
            if concentrator_tunnel_ip:
                wgc_if.ip4_addr = ipaddress.IPv4Interface(str(concentrator_tunnel_ip))
            
            prerequisites = (is_enabled, concentrator_endpoint, concentrator_password, concentrator_port, concentrator_tunnel_ip)
            
            if [x for x in prerequisites if x is None or x == '' or x is False]:
                self.logger.debug("ON RETURN")
                continue
            
            wgc.WireGuardConcentratorClient.WireGuardConcentratorClientStartup(wgc_if._wgif_idx, wgc_if._name, concentrator_port, concentrator_tunnel_ip)
            
        return True
        
    def apply_running_config(self, name, config, change_set):
        self.logger.debug('Apply RUNNING config for Wireguard Concentrator')
        self.logger.error(f"Change_set: {change_set}")
        
        c_enabled = False
        concentrator_endpoint = None
        concentrator_port = None
        concentrator_password = None
        ipv4_prefix = None

        self.logger.error("Config {}".format(config))
        if config:
            c_enabled = config.get('enable', False)
            ipv4_prefix = config.get('ipv4-prefix', None)
            ifname = config.get('name', None)
            concentrator_endpoint = config.get('concentrator-endpoint', None)
            concentrator_password = config.get('concentrator-password', None)
            concentrator_port = config.get('concentrator-port', None)

            self.logger.debug(f"{ifname} {c_enabled} {ipv4_prefix} {concentrator_endpoint} {concentrator_password} {concentrator_port}")
        
        if change_set:
            self.logger.debug("APPLYING RUNNING FOR WIREGUARD INTERFACE - {}".format(name))
            # self.logger.debug('change_set {} ' .format(change_set))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))

            wg_if = next((a for a in self._wgc_interfaces if a.name == name), None)
            if not wg_if:
                wg_if = wireguardif.WireGuardVPNInterface(name, logger=self.logger, l_port=concentrator_port,
                                                l_endpoint=concentrator_endpoint, enable=c_enabled, is_concentrator=True)
                self._wgc_interfaces.append(wg_if)

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if wg_if:
                            self.logger.error("There is WireGuard interface object for {} ?!?!".format(name))
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if wg_if:
                            self._wgc_interfaces.remove(wg_if)
                            if not self._wgc_interfaces:
                                #Restore systemd-resolved
                                self._restore_systemd_resolved()
                            del wg_if
                        return
            
            if 'concentrator-port' in ch_set_keys:
                oper = change_set.get('concentrator-port').get('operation', None )
                local_port = None

                self.logger.error('Local port is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    local_port = change_set.get('concentrator-port').get('new_value', None)

                wg_if.l_port = int(local_port)
                
            if 'concentrator-endpoint' in ch_set_keys:
                oper = change_set.get('concentrator-endpoint').get('operation', None )
                local_endpoint = None

                self.logger.error('Local endpoint is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    local_endpoint = change_set.get('concentrator-endpoint').get('new_value', None)

                l_local, l_localv6 = self._get_local_wireguard_endpoint_ip(local_endpoint)
                
                if l_local:
                    wg_if.l_endpoint = l_local
                    self.logger.error('Local endpoint configured {}' .format(l_local))
                if l_localv6:
                    wg_if.l_endpoint_v6 = l_localv6
                    self.logger.error('Local endpoint configured {}' .format(l_localv6))
                    
            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                c_enable = False

                self.logger.error('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('enable').get('new_value', False) and \
                            'true' in str(change_set.get('enable').get('new_value', False)).lower():
                        c_enable = True
                        # self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 1)
                    else:
                        # self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 0)
                        pass
                    
                wg_if.enable = c_enable
                
                if c_enable:
                    self.logger.debug(f"wg_if._wgif_idx {wg_if._wgif_idx} wg_if._name {wg_if._name} concentrator_port {concentrator_port} ")
                    wgc.WireGuardConcentratorClient.WireGuardConcentratorClientStartup(wg_if._wgif_idx, wg_if._name, concentrator_port, ipv4_prefix)
                
            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                c_ip4_addr = None

                self.logger.error('IPv4 prefix is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_ip4_addr = change_set.get('ipv4-prefix').get('new_value', None)
                    if c_ip4_addr:
                        c_ip4_addr = ipaddress.IPv4Interface(str(c_ip4_addr))
                    
                wg_if.ip4_addr = c_ip4_addr
                
            if 'domain-name' in ch_set_keys:
                oper = change_set.get('domain-name').get('operation', None)
                c_d_name = None

                self.logger.error('Domain name is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_d_name = change_set.get('domain-name').get('new_value', None)
                
                # set dns name in /etc/dnsmasq.conf
                wg_if.d_name = c_d_name

            for key in ch_set_keys:
                if 'access-list/' in key:
                    # Identify the exact action (permit/deny) from the key
                    self.logger.error(f"KEY {key}")
                    action = 'permit' if 'permit' in key else 'deny'
                    oper = change_set.get(key, {}).get('operation', None)

                    self.logger.error(f'Access-list changed - operation {oper}')

                    if oper in ['MODIFIED', 'CREATED']:
                        c_a_list = change_set.get(key, {}).get('new_value', None)
                        if c_a_list:
                            if 'rule' in c_a_list:
                                wg_if.a_list = {action: c_a_list}
                                self.logger.error(f'Applied rule: {c_a_list} for action: {action}')

                    elif oper == 'DELETED':
                        c_a_list = change_set.get(key, {}).get('old_value', None)
                        if c_a_list:
                            if 'rule' in c_a_list:
                                c_a_list += ' delete'
                                wg_if.a_list = {action: c_a_list}
                                self.logger.error(f'Applied deletion rule: {c_a_list} for action: {action}')

    def post_apply_config(self, change_set):
        self.logger.error(f"POST APPLY- change_set {change_set}")
        config_indexed = self.get_wireguard_concentrators(startup=True)
        self.logger.error(f"STARTUP {config_indexed}")

    
    def apply_config(self, change_set=None, apply_startup=False):
        self.logger.debug("Apply Wireguard Concentrator configuration")

        if apply_startup:
            self.apply_startup_config()
        else:
            config_indexed = self.get_wireguard_concentrators(startup=False)
            if not change_set:
                self.logger.error(" There are no changes in change set")
                return config_indexed
            
            concentrator_changes = {}
            for change in change_set:
                if not 'operation' in change:
                    self.logger.error('Malformed change set')
                    continue
                
                name = str(change['xpath'])
                name = name[name.find("wg-concentrator[name='") + len("wg-concentrator[name='"):]
                name = name[:name.find("]") - 1]
                if name not in concentrator_changes:
                    concentrator_changes[name] = {}
                
                concentrator_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change
            
            for name, xpath_dict in concentrator_changes.items():
                if name in config_indexed:
                    self.apply_running_config(name, config_indexed[name], xpath_dict)
                else:
                    self.apply_running_config(name, None, xpath_dict)

        return True

    # params to set: concentrator-endpoint, concentrator-password, ipv4-prefix, enable, static-range, dynamic-range, domain-name, static-routes;
    def get_wireguard_concentrator_interfaces(self):
        return self.get_running_config(full_path=False)

    # params to set: concentrator-endpoint, concentrator-password, ipv4-prefix, enable, static-ip-range, dynamic-ip-range, domain-name, static-routes;
    def set_wireguard_concentrator_param(self, ifname, param, value):
        self.set_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, ifname, param), value)

    def unset_wireguard_concentrator_param(self, ifname, param):
        self.delete_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, ifname, param))

    def del_wireguard_concentrator(self, name):
        self.delete_item("{}/wg-concentrator[name='{}']".format(self.namespace, name))

    def get_wireguard_concentrator_param(self, name, param, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, name, param))
        except:
            return default

    def get_wireguard_concentrator_params(self, name, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']".format(self.namespace, name))
        except:
            return default

    def set_wireguard_concentrator_static_clients(self, ifname, hostname, param, value):
        return self.set_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']/{}".format(self.namespace, ifname, hostname, param), value)

    def unset_wireguard_concentrator_static_clients(self, ifname, hostname, param):
        return self.delete_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']/{}".format(self.namespace, ifname, hostname, param))

    def del_wireguard_concentrator_static_clients(self, ifname, hostname):
        return self.delete_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']".format(self.namespace, ifname, hostname))

    def get_wireguard_concentrator_static_client(self, name, hostname, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']".format(self.namespace, name, hostname))
        except:
            return default

    def get_wireguard_concentrator_static_clients(self, name, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']/static-clients".format(self.namespace, name))
        except:
            return default

    def set_wireguard_concentrator_external_peers(self, ifname, endpoint, param, value):
        return self.set_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param), value)

    def unset_wireguard_concentrator_external_peers(self, ifname, endpoint, param):
        return self.delete_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param))

    def del_wireguard_concentrator_external_peers(self, ifname, endpoint):
        return self.delete_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint))

    def get_wireguard_concentrator_external_peer(self, name, endpoint, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']".format(self.namespace, name, endpoint))
        except:
            return default

    def get_wireguard_concentrator_external_peers(self, name, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/external-peers".format(self.namespace, name))
        except:
            return default

    def set_wireguard_concentrator_acl(self, ifname, acl_type, acl_id, rule):
        return self.set_item("{}/wg-concentrator[name='{}']/access-list/{}[id='{}']/rule".format(self.namespace, ifname, acl_type, acl_id), rule)

    def unset_wireguard_concentrator_acl(self, ifname, acl_type, acl_id):
        self.delete_item("{}/wg-concentrator[name='{}']/access-list/{}[id='{}']".format(self.namespace, ifname, acl_type, acl_id))

    def get_wireguard_concentrator_acl(self, ifname, acl_type, acl_id):
        return self.get_item("{}/wg-concentrator[name='{}']/access-list/{}[id='{}']/rule".format(self.namespace, ifname, acl_type, acl_id))

    def get_wireguard_concentrator_acls(self, ifname, acl_type):
        config = self.get_running_config(full_path=False)
        config_indexed = {}

        if 'wg-concentrator' not in config:
            return config_indexed

        for concentrator in config['wg-concentrator']:
            if concentrator['name'] == ifname and 'access-list' in concentrator:
                # Check for the correct ACL type (permit/deny)
                if acl_type in concentrator['access-list']:
                    for rule in concentrator['access-list'][acl_type]:
                        config_indexed[rule['id']] = rule
                break

        return config_indexed

    def get_wireguard_concentrator_next_acl_id(self, ifname, acl_type):
        acl_rules = self.get_wireguard_concentrator_acls(ifname, acl_type)  # Get existing ACL rules indexed by id
        existing_ids = [int(rule_id) for rule_id in acl_rules.keys()]  # Extract existing IDs

        # Find the next available ID
        next_id = 1
        while next_id in existing_ids:
            next_id += 1

        return next_id

    def find_wireguard_concentrator_acl_rule(self, ifname, acl_type, rule):
        acl_entries = self.get_wireguard_concentrator_acls(ifname, acl_type)

        # Iterate through the ACL entries and search for the specified rule
        for acl_id, acl_entry in acl_entries.items():
            if acl_entry.get('rule') == rule:
                return acl_id

        return None

    def _get_local_wireguard_endpoint_ip(self, endpoint):
        n_local = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv4Address) else None
        n_localv6 = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv6Address) else None

        return (n_local, n_localv6)

    def get_wireguard_interfaces_from_vpp(self):
        vpp_if = vppif.VppApiInterface('show-wireguard')

        ret = vpp_if.wg_get_interfaces()
        intfs_list = []
        for iface in ret:
            intfs_list.append(iface)

        return intfs_list

    def get_wireguard_peers_from_vpp(self):
        vpp_if = vppif.VppApiInterface('show-wireguard')
        ret = vpp_if.wg_get_peers()
        peers_list = []
        for peer in ret:
            interface_name = vpp_if.get_ifname_by_ifidx(peer['wgif'])
            allowed_ips = [str(ip) for ip in peer['allowed_ips']]
            peer_endpoint = str(ipaddress.IPv4Address(peer['peer_endpoint']))
            peers_list.append({
                'interface': interface_name,
                'public_key': peer['public_key'],
                'allowed_ips': allowed_ips,
                'peer_endpoint': peer_endpoint,
                'peer_port': peer['peer_port']
            })
        return peers_list

    def find_unused_ips(self, start_address, end_address, used_ips):
        def format(s, e):
            s = str(ipaddress.IPv4Address(s))
            e = str(ipaddress.IPv4Address(e))
            return f'{s} - {e}'

        # Convert the tuple of used IPs to a set for faster lookups
        used_ips_set = set(ip[0] for sublist in used_ips if sublist for ip in sublist)

        unused_ips = []
        ips = []
        skip = 0
        for ip in range(start_address + 1, end_address):
            if ip not in used_ips_set:
                unused_ips.append(ip)
        unused_ips.sort()

        if unused_ips[-1] - unused_ips[0] + 1 == len(unused_ips):
            return [format(unused_ips[0], unused_ips[-1])]
        for idx, ip in enumerate(unused_ips):
            if skip > 0:
                skip -= 1
                continue
            i = ip + 1
            if len(unused_ips[idx:]) == 1:
                ips.append(str(ipaddress.IPv4Address(ip)))
                break
            if unused_ips[-1] - ip + 1 == len(unused_ips[idx:]):
                ips.append(format(ip, unused_ips[-1]))
                break
            for n_ip in unused_ips[idx+1:]:
                if n_ip == i:
                    i += 1
                elif i - ip > 1:
                    ips.append(format(ip, i - 1))
                    skip = i - ip - 1
                    break
                else:
                    ips.append(str(ipaddress.IPv4Address(ip)))
                    break

        return ips

    def get_wireguard_available_ips(self, intf_name):
        edge_db = db.EdgilityDatabase()

        taken_ips = []
        tables = edge_db.get_all_tables('wireguard')
        for table in tables:
            if table.startswith(('dynamic_clients_', 'static_clients_')):
                taken_ips.append(edge_db.fetch_column_entries(table, 'tunnel_ip'))

        tunnel_ip = self.get_wireguard_concentrator_param(intf_name, 'ipv4-prefix')
        if not tunnel_ip:
            return None

        net_info = ipaddress.IPv4Network(f"{tunnel_ip}", strict=False)
        net_addr = int(net_info.network_address)
        available_hosts = net_info.num_addresses

        taken_ips.insert(0, [((int(ipaddress.ip_address(tunnel_ip.split('/')[0]))),)])
        free_ips = self.find_unused_ips(net_addr, net_addr + available_hosts - 1, taken_ips)

        return free_ips

    def get_wireguard_dynamic_peers_from_db(self, inft_name):
        edge_db = db.EdgilityDatabase()

        try:
            data = edge_db.fetch_all_rows_from_table('dynamic_clients_' + inft_name)
            if data:
                converted_data = [
                    {
                        'public_key': row[0],
                        'hostname': row[1],
                        'endpoint': row[2],
                        'tunnel_ip': str(ipaddress.IPv4Address(row[3])),
                        'password': row[5]
                    } for row in data
                ]
                return converted_data
        except:
            return None

    def get_wireguard_static_peers_from_db(self, inft_name):
        edge_db = db.EdgilityDatabase()

        try:
            data = edge_db.fetch_all_rows_from_table('static_clients_' + inft_name)
            if data:
                converted_data = [
                    {
                        'hostname': row[0],
                        'tunnel_ip': str(ipaddress.IPv4Address(int(row[1])))
                    } for row in data
                ]
                return converted_data
        except:
            return None

    def _restore_systemd_resolved(self):
        utils.execute_shell_command(f'sudo systemctl enable systemd-resolved.service')
        utils.execute_shell_command(f'sudo systemctl start systemd-resolved.service')
        utils.execute_shell_command(f'sudo ln -sf /run/systemd/resolve/stub-resolv.conf /etc/resolv.conf')
        utils.execute_shell_command(f'sudo systemctl disable dnsmasq.service')
